#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cengal.code_flow_control.chained_flow.versions.v_1.chained_flow import *
from random import randint

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def random_int(a, b):
    if a < -20:
        raise ChainLinkFailed('Too small!')
    result = randint(a, b)
    return result


def int_div(a, b):
    result = a / b
    return result


def int_square(value):
    result = value * value
    return result


def user_input(is_ok__context_holder=None):
    if is_ok__context_holder:
        uinput = input('Enter something:')
        str_len = len(uinput)
        if str_len < 10:
            raise ChainLinkFailed('Too short!')
        result = uinput
        is_ok__context_holder.push_result(True, uinput)
        return result


def print_result(result_block_id, is_ok__context_holder=None):
    if is_ok__context_holder:
        print('RESULT FOR "{}":'.format(is_ok__context_holder._context_holder_id),
              is_ok__context_holder.read_block_result_link(result_block_id).result)
    else:
        print('ERROR IN "{}": Can\'t compute!'.format(is_ok__context_holder._context_holder_id))
        print(str(is_ok__context_holder))


@link__function__simple
def random_int_d(a, b):
    if a < -20:
        raise ChainLinkFailed('Too small!')
    result = randint(a, b)
    return result


@link__function__simple
def int_div_d(a, b):
    result = a / b
    return result


@link__function__simple
def int_square_d(value):
    result = value * value
    return result


@link__function
def user_input_d(is_ok__context_holder=None):
    if is_ok__context_holder:
        uinput = input('Enter something:')
        str_len = len(uinput)
        if str_len < 10:
            raise ChainLinkFailed('Too short!')
        result = uinput
        is_ok__context_holder.push_result(True, uinput)
        return result


@chain_reader__function
def print_result_d(result_block_id, is_ok__context_holder=None):
    if is_ok__context_holder:
        print('RESULT:', is_ok__context_holder.read_block_result_link(result_block_id).result)
    else:
        print('Error: Can\'t compute!')
        print(str(is_ok__context_holder))


def with_chain_example(raise_exceptions=False):
    """
    Console output:

========================
=== WITH-CHAIN-RAISE-EXCEPTIONS: ===


WITH raise_exceptions:


>> loop iteration #0 START

Ops: ZeroDivisionError

Oh! There are errors in Main Loop! Interesting!

{{CONTEXT_HOLDER_ID(Test Holder): CONTEXT_HOLDER_INFO(Some holder)}:[
(index(0), depth(1), ID(1 str), INFO(None), RESULT((False, ISOK INTERNAL RESULT. EXTERNAL EXCEPTION: <class 'ZeroDivisionError'> division by zero ))),
(index(1), depth(0), ID(main loop), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BUILT-IN EXCEPTION: BAD HISTORY IMPORT EXCEPTION)))
]}



========================
========================
========================



========================
=== WITH-CHAIN: ===


WITHOUT raise_exceptions:


>> loop iteration #0 START

Ops: ZeroDivisionError
Sum was not computed because of "Did not fit criteria because of "14" in 2 block" in "Sum" block.
Mul: 120.0
GlobalSum of (Mul, Sum, 0): 126.0
GlobalSum2 of (Mul, Sum, 0): 126.0

>> loop iteration #0 END
{{CONTEXT_HOLDER_ID(Test Holder): CONTEXT_HOLDER_INFO(Some holder)}:[
(index(0), depth(0), ID(0), INFO(None), RESULT((True, 6))),
(index(1), depth(0), ID(1 str), INFO(None), RESULT((False, ISOK INTERNAL RESULT. EXTERNAL EXCEPTION: <class 'ZeroDivisionError'> division by zero ))),
(index(2), depth(0), ID(2), INFO(None), RESULT((True, 14))),
(index(3), depth(0), ID(2 str), INFO(None), RESULT((False, ISOK INTERNAL RESULT. EXTERNAL EXCEPTION: <class 'ZeroDivisionError'> division by zero ))),
(index(4), depth(0), ID(Hello), INFO(None), RESULT((True, 20.0))),
(index(5), depth(0), ID(Sum), INFO(None), RESULT((False, Did not fit criteria because of "14" in 2 block))),
(index(6), depth(0), ID(Mul), INFO(None), RESULT((True, 120.0))),
(index(7), depth(0), ID(print1), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT))),
(index(8), depth(0), ID(print2), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT))),
(index(9), depth(0), ID(GlobalSum), INFO(None), RESULT((True, 126.0))),
(index(10), depth(0), ID(GlobalSum2), INFO(None), RESULT((True, 126.0)))
]}


Everything is fine: result fit criteria of current reader block: "ResultType(CriteriaType.forbidden, {'1'})"!


Everything is fine: result fit criteria of current reader block: "ResultType(CriteriaType.forbidden, {'2'})"!


>> loop iteration #1 START

Ops: ZeroDivisionError
Sum was not computed because of "Did not fit criteria because of "14" in 2 block" in "Sum" block.
Mul: 120.0
GlobalSum of (Mul, Sum, 0): 126.0
GlobalSum2 of (Mul, Sum, 0): 126.0

>> loop iteration #1 END
{{CONTEXT_HOLDER_ID(Test Holder): CONTEXT_HOLDER_INFO(Some holder)}:[
(index(0), depth(0), ID(0), INFO(None), RESULT((True, 6))),
(index(1), depth(0), ID(1 str), INFO(None), RESULT((False, ISOK INTERNAL RESULT. EXTERNAL EXCEPTION: <class 'ZeroDivisionError'> division by zero ))),
(index(2), depth(0), ID(2), INFO(None), RESULT((True, 14))),
(index(3), depth(0), ID(2 str), INFO(None), RESULT((True, 54.0))),
(index(4), depth(0), ID(Hello), INFO(None), RESULT((True, 20.0))),
(index(5), depth(0), ID(Sum), INFO(None), RESULT((False, Did not fit criteria because of "14" in 2 block))),
(index(6), depth(0), ID(Mul), INFO(None), RESULT((True, 120.0))),
(index(7), depth(0), ID(print1), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT))),
(index(8), depth(0), ID(print2), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT))),
(index(9), depth(0), ID(GlobalSum), INFO(None), RESULT((True, 126.0))),
(index(10), depth(0), ID(GlobalSum2), INFO(None), RESULT((True, 126.0)))
]}


Everything is fine: result fit criteria of current reader block: "ResultType(CriteriaType.forbidden, {'1'})"!


Everything is NOT fine: did not fit default criteria set in context constructor: "ResultType(CriteriaType.forbidden, {'2'})"!


Oh! There are errors in Main Loop! Interesting!

{{CONTEXT_HOLDER_ID(Test Holder): CONTEXT_HOLDER_INFO(Some holder)}:[
(index(0), depth(1), ID(1 str), INFO(None), RESULT((False, ISOK INTERNAL RESULT. EXTERNAL EXCEPTION: <class 'ZeroDivisionError'> division by zero ))),
(index(1), depth(1), ID(Sum), INFO(None), RESULT((False, Did not fit criteria because of "14" in 2 block))),
(index(2), depth(1), ID(print1), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT))),
(index(3), depth(1), ID(print2), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT))),
(index(4), depth(0), ID(main loop), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BUILT-IN EXCEPTION: BAD HISTORY IMPORT EXCEPTION)))
]}


========================

    :param raise_exceptions:
    :return:
    """
    if raise_exceptions:
        print()
        print('WITH raise_exceptions:')
        print()
    else:
        print()
        print('WITHOUT raise_exceptions:')
        print()

    function_context = Chain('Test Holder', 'Some holder', ResultType(CriteriaType.optional, set()))

    with link(function_context, 'main loop'):
        if function_context:
            for index in range(10):
                print()
                print('>> loop iteration #{} START'.format(index))
                print()

                context = Chain('Test Holder', 'Some holder', ResultType(CriteriaType.forbidden, {'1 str'}),
                                raise_exceptions=raise_exceptions)

                with link(context, 0, None, ResultType(CriteriaType.any, None)):
                    if context:
                        result = 2 * 3
                        context.push_result(True, result)

                try:
                    with link(context, '1 str', None, ResultType(CriteriaType.any, None)):
                        if context:
                            try:
                                result = 54 / 0  # if raise_exceptions - it will break Main Loop
                                context.push_result(True, result)
                            except ZeroDivisionError:
                                print('Ops: ZeroDivisionError')
                                raise
                except ZeroDivisionError:
                    print('THIS LINE WILL NOT BE EXECUTED!')

                with link(context, 2):
                    if context:
                        result = 14
                        context.push_result(True, result)
                    else:
                        raise ChainLinkFailed('Something went wrong.')

                try:
                    with link(context, '2 str'):
                        if context:
                            result = 54 / index
                            context.push_result(True, result)
                except ZeroDivisionError:
                    print('THIS LINE WILL NOT BE EXECUTED!')

                with link(context, 'Hello', None, ResultType(CriteriaType.needed, set())):
                    if context:
                        result = 100 / 5
                        context.push_result(True, result)

                with link(context, 'Sum', None, ResultType(CriteriaType.forbidden, {0, 2})):
                    if context:
                        result = context.read_block_result_link(0).result + context.read_block_result_link(2).result
                        context.push_result(True, result)
                    else:
                        context.push_result(False, 'Did not fit criteria because of "{}" in 2 block'.format(
                            context.read_block_result_link(2).result))

                with link(context, 'Mul', None, ResultType(CriteriaType.needed, {0, 'Hello'})):
                    if context:
                        result = context.read_block_result_link(0).result * context.read_block_result_link('Hello').result
                        context.push_result(True, result)

                with link(context, 'print1', None, ResultType(CriteriaType.needed, {'Sum'})):
                    if context:
                        print('Sum: {}'.format(context.read_block_result_link('Sum').result))
                    else:
                        print('Sum was not computed because of "{}" in "Sum" block.'.format(
                            context.read_block_result_link('Sum').result
                        ))

                with link(context, 'print2', None, ResultType(CriteriaType.needed, {'Mul'})):
                    if context:
                        print('Mul: {}'.format(context.read_block_result_link('Mul').result))

                with link(context, 'GlobalSum', None, ResultType(CriteriaType.needed, set())):
                    if context:
                        interesting_results = {0, 'Sum', 'Mul'}
                        global_sum = 0
                        for interesting_result in interesting_results:
                            if context.read_block_result_link(interesting_result):
                                global_sum += context.read_block_result_link(interesting_result).result

                        print('GlobalSum of ({}): {}'.format(', '.join(str(x) for x in interesting_results), global_sum))
                        context.push_result(True, global_sum)

                with link(context, 'GlobalSum2'):
                    interesting_results = {0, 'Sum', 'Mul'}
                    global_sum = 0
                    for interesting_result in interesting_results:
                        if context.read_block_result_link(interesting_result):
                            global_sum += context.read_block_result_link(interesting_result).result

                    print('GlobalSum2 of ({}): {}'.format(', '.join(str(x) for x in interesting_results), global_sum))
                    context.push_result(True, global_sum)

                print()
                print('>> loop iteration #{} END'.format(index))
                print(str(context))
                print()

                with chain_reader(context):
                    if context:
                        print()
                        print('Everything is fine: result fit criteria of current reader block: '
                              '"ResultType(CriteriaType.forbidden, {\'1\'})"!')
                        print()
                    else:
                        context.raise_bad_blocks()

                with chain_reader(context, ResultType(CriteriaType.forbidden, {'2 str'})):
                    if context:
                        print()
                        print('Everything is fine: result fit criteria of current reader block: '
                              '"ResultType(CriteriaType.forbidden, {\'2\'})"!')
                        print()
                    else:
                        print()
                        print('Everything is NOT fine: did not fit default criteria set in context constructor: '
                              '"ResultType(CriteriaType.forbidden, {\'2\'})"!')
                        print()
                        context.raise_bad_blocks()

    with chain_reader(function_context):
        if function_context:
            print()
            print('Everything is fine!')
            print()
        else:
            print()
            print('Oh! There are errors in Main Loop! Interesting!')
            print()
            print(str(function_context))
            print()


def func_decorator_chain_example():
    """
    Console output:

========================
=== FUNC-DECORATOR-COMPLICATED-CHAIN: ===

Enter something:Hello
RESULT: 625.0
RESULT: ISOK INTERNAL RESULT. BUILT-IN EXCEPTION: ChainLinkFailed (Too short!)


========================

    :return:
    """
    cont = Chain('Test Holder', 'Some holder', ResultType(CriteriaType.any, None))

    user_input_d(is_ok__context_holder=cont, is_ok__block_id='User Input')

    result_block_id = 'Result'

    # Will fail: -40 < -20.
    int_square_d(
        int_div_d(
            random_int_d(-40, 0, is_ok__context_holder=cont),  # -40 < -20
            random_int_d(3, 4, is_ok__context_holder=cont),
            is_ok__context_holder=cont),
        is_ok__context_holder=cont, is_ok__block_id=result_block_id,
        is_ok__block_results_criteria=ResultType(CriteriaType.not_successful, {result_block_id}),
        is_ok__ignore_block_result_criteria=IgnoreBlockResultCriteriaType.ignore_if_failed)

    # Will be successful.
    int_square_d(
        int_div_d(
            random_int_d(100, 100, is_ok__context_holder=cont),  # 100
            random_int_d(4, 4, is_ok__context_holder=cont), is_ok__context_holder=cont  # 4
        ),  # 25
        is_ok__context_holder=cont,
        is_ok__block_id=result_block_id,
        is_ok__block_results_criteria=ResultType(CriteriaType.not_successful, {result_block_id}),
        is_ok__ignore_block_result_criteria=IgnoreBlockResultCriteriaType.ignore_if_failed
    )  # 625

    # Will fail: 'Result' already successfully computed
    # Also will fail in stand alone mode: division by zero.
    int_1 = random_int_d(-10, 0, is_ok__context_holder=cont)
    int_2 = random_int_d(0, 0, is_ok__context_holder=cont)  # 0 (zero)
    div_res = int_div_d(int_1, int_2, is_ok__context_holder=cont)  # value / 0
    int_square_d(div_res, is_ok__context_holder=cont, is_ok__block_id=result_block_id
                 , is_ok__block_results_criteria=ResultType(CriteriaType.not_successful, {result_block_id})
                 , is_ok__ignore_block_result_criteria=IgnoreBlockResultCriteriaType.ignore_if_failed)

    # Will print "625"
    print_result_d(result_block_id, is_ok__context_holder=cont)

    # Will print user input or error (exception info) with text "Too short!"
    print_result_d('User Input', is_ok__context_holder=cont)


def func_runner_chain_complicated_example():
    """
    Console output:

========================
=== FUNC-RUNNER-COMPLICATED-CHAIN: ===

Enter something:World
RESULT FOR "Test Holder": 625.0
RESULT FOR "Test Holder": ISOK INTERNAL RESULT. BUILT-IN EXCEPTION: ChainLinkFailed (Too short!)


========================

    :return:
    """
    cont = Chain('Test Holder', 'Some holder', ResultType(CriteriaType.any, None))

    cr = ChainUniRunner(globals(), cont, True)  # Simple-Mode - for any functions

    ChainUniRunner(globals(), cont)('User Input').user_input()  # Normal-Mode - for specially prepared functions

    result_block_id = 'Result'

    # Will fail: -40 < -20.
    # Will return None if you'll try to assign final result to the variable
    cr(result_block_id, None, ResultType(CriteriaType.not_successful, {result_block_id}),
       IgnoreBlockResultCriteriaType.ignore_if_failed
       ).int_square(
        cr().int_div(
            cr().random_int(-40, 0),  # -40 < -20
            cr().random_int(3, 4)))

    # Will be successful.
    # Will return 625 if you'll try to assign final result to the variable
    cr(result_block_id,
       block_results_criteria=ResultType(CriteriaType.not_successful, {result_block_id}),
       ignore_block_result_criteria=IgnoreBlockResultCriteriaType.ignore_if_failed
       ).int_square(
        cr().int_div(
            cr().random_int(100, 100),  # 100
            cr().random_int(4, 4)  # 4
        )  # 25
    )  # 625

    # Will fail: 'Result' already successfully computed
    # Also will fail in stand alone mode: division by zero.
    # Will return None if you'll try to assign final result to the variable
    int_1 = cr().random_int(-10, 0)
    int_2 = cr().random_int(0, 0)  # 0 (zero)
    div_res = cr().int_div(int_1, int_2)  # value / 0
    cr(result_block_id,
       block_results_criteria=ResultType(CriteriaType.not_successful, {result_block_id}),
       ignore_block_result_criteria=IgnoreBlockResultCriteriaType.ignore_if_failed
       ).int_square(div_res)

    # Will print "625"
    ChainReaderRunner(globals(), cont)().print_result(result_block_id)

    # Will print user input or error (exception info) with text "Too short!"
    crr = ChainReaderRunner(globals(), cont)
    crr().print_result('User Input')


def func_runner_chain_simple_example():
    """
    Console output:

========================
=== FUNC-RUNNER-SIMPLE-CHAIN: ===

RESULT FOR "Test Holder Number 1": 625.0

ERROR IN "Test Holder Number 2": Can't compute!
{{CONTEXT_HOLDER_ID(Test Holder Number 2): CONTEXT_HOLDER_INFO(Will fail)}:[
(index(0), depth(0), ID(<function random_int at 0x000001D1D8A4EE18>), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BUILT-IN EXCEPTION: ChainLinkFailed (Too small!)))),
(index(1), depth(0), ID(<function random_int at 0x000001D1D8A4EE18>), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT))),
(index(2), depth(0), ID(<function int_div at 0x000001D1D8A4EEA0>), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT))),
(index(3), depth(0), ID(Result), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT)))
]}

ERROR IN "Test Holder Number 3": Can't compute!
{{CONTEXT_HOLDER_ID(Test Holder Number 3): CONTEXT_HOLDER_INFO(Every function should return number >= 0. Will fail.)}:[
(index(0), depth(0), ID(<function random_int at 0x000001D1D8A4EE18>), INFO(None), RESULT((True, 100))),
(index(1), depth(0), ID(<function random_int at 0x000001D1D8A4EE18>), INFO(None), RESULT((False, -3))),
(index(2), depth(0), ID(<function int_div at 0x000001D1D8A4EEA0>), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT))),
(index(3), depth(0), ID(Result), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT)))
]}

ERROR IN "Test Holder Number 4": Can't compute!
{{CONTEXT_HOLDER_ID(Test Holder Number 4): CONTEXT_HOLDER_INFO(Every function should return even number. Will fail.)}:[
(index(0), depth(0), ID(<function random_int at 0x000001D1D8A4EE18>), INFO(None), RESULT((True, 100))),
(index(1), depth(0), ID(<function random_int at 0x000001D1D8A4EE18>), INFO(None), RESULT((True, 4))),
(index(2), depth(0), ID(<function int_div at 0x000001D1D8A4EEA0>), INFO(None), RESULT((False, 25.0))),
(index(3), depth(0), ID(Result), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT)))
]}


========================

    :return:
    """
    # ============================
    # TEST 1:
    cont = Chain('Test Holder Number 1', 'Will be successful', ResultType(CriteriaType.optional, set()))
    cr = ChainUniRunner(globals(), cont, True)  # Simple-Mode - for any functions

    # Will be successful.
    # Will return 625 if you'll try to assign final result to the variable
    cr('Result').int_square(
        cr().int_div(
            cr().random_int(100, 100),  # 100
            cr().random_int(4, 4)  # 4
        )  # 25
    )  # 625

    # Will print "625"
    ChainReaderRunner(globals(), cont)().print_result('Result')

    # ============================
    print()

    # ============================
    # TEST 2:
    cont = Chain('Test Holder Number 2', 'Will fail', ResultType(CriteriaType.optional, set()))
    cr = ChainUniRunner(globals(), cont, True)  # Simple-Mode - for any functions

    # Will fail: -40 < -20.
    # Will return None if you'll try to assign final result to the variable
    cr('Result').int_square(  # Will not be computed
        cr().int_div(  # Will not be computed
            cr().random_int(-40, 0),  # -40 < -20. EXCEPTION ChainLinkFailed ("Too small!")
            cr().random_int(3, 4)))  # Will not be computed

    # Will print an error
    ChainReaderRunner(globals(), cont)().print_result('Result')

    # ============================
    print()

    # ============================
    # TEST 3:
    cont = Chain('Test Holder Number 3', 'Every function should return number >= 0. Will fail.',
                 ResultType(CriteriaType.optional, set()))
    cr = ChainUniRunner(globals(), cont, True, lambda result: result >= 0)  # Simple-Mode - for any functions

    # Will fail: random_int(-5, -1) < 0.
    # Will return None if you'll try to assign final result to the variable
    cr('Result').int_square(
        cr().int_div(
            cr().random_int(100, 100),  # 100
            cr().random_int(-5, -1)  # result < 0. Result state is FAIL (push_result(False, ...))
        )  # Will not be computed
    )  # Will not be computed

    # Will print an error
    ChainReaderRunner(globals(), cont)().print_result('Result')

    # ============================
    print()

    # ============================
    # TEST 4:
    cont = Chain('Test Holder Number 4', 'Every function should return even number. Will fail.',
                 ResultType(CriteriaType.optional, set()))
    cr = ChainUniRunner(globals(), cont, True, lambda result: result % 2 == 0)  # Simple-Mode - for any functions

    # Will fail: 25 is odd.
    # Will return None if you'll try to assign final result to the variable
    int_1 = cr().random_int(100, 100)  # 100
    int_2 = cr().random_int(4, 4)  # 4
    div_res = cr().int_div(int_1, int_2)  # 25. Result is odd. Result state is FAIL (push_result(False, 25))
    cr('Result').int_square(div_res)  # Will not be computed

    # Will print an error
    ChainReaderRunner(globals(), cont)().print_result('Result')


def func_call_runner_chain_complicated_example():
    """
    Console output:

========================
=== FUNC-CALL-RUNNER-COMPLICATED-CHAIN: ===

Enter something:Hello World
RESULT FOR "Test Holder": 625.0
RESULT FOR "Test Holder": Hello World


========================

    :return:
    """
    cont = Chain('Test Holder', 'Some holder', ResultType(CriteriaType.any, None))

    cr = ChainUniCallRunner(cont, True)  # Simple-Mode - for any functions

    ChainUniCallRunner(cont)('User Input')(user_input)  # Normal-Mode - for specially prepared functions

    result_block_id = 'Result'

    # Will fail: -40 < -20.
    # Will return None if you'll try to assign final result to the variable
    cr(result_block_id, None, ResultType(CriteriaType.not_successful, {result_block_id}),
       IgnoreBlockResultCriteriaType.ignore_if_failed)(
        int_square,
        cr()(
            int_div,
            cr()(random_int, -40, 0),  # -40 < -20
            cr()(random_int, 3, 4)))

    # Will be successful.
    # Will return 625 if you'll try to assign final result to the variable
    cr(result_block_id,
       block_results_criteria=ResultType(CriteriaType.not_successful, {result_block_id}),
       ignore_block_result_criteria=IgnoreBlockResultCriteriaType.ignore_if_failed)(
        int_square,
        cr()(
            int_div,
            cr()(random_int, 100, 100),  # 100
            cr()(random_int, 4, 4)  # 4
        )  # 25
    )  # 625

    # Will fail: 'Result' already successfully computed
    # Also will fail in stand alone mode: division by zero.
    # Will return None if you'll try to assign final result to the variable
    int_1 = cr()(random_int, -10, 0)
    int_2 = cr()(random_int, 0, 0)  # 0 (zero)
    div_res = cr()(int_div, int_1, int_2)  # value / 0
    cr(result_block_id,
       block_results_criteria=ResultType(CriteriaType.not_successful, {result_block_id}),
       ignore_block_result_criteria=IgnoreBlockResultCriteriaType.ignore_if_failed
       )(int_square, div_res)

    # Will print "625"
    ChainReaderCallRunner(cont)()(print_result, result_block_id)

    # Will print user input or error (exception info) with text "Too short!"
    crr = ChainReaderCallRunner(cont)
    crr()(print_result, 'User Input')


def func_call_runner_chain_simple_example():
    """
    Console output:

========================
=== FUNC-CALL-RUNNER-SIMPLE-CHAIN: ===

RESULT FOR "Test Holder Number 1": 625.0

ERROR IN "Test Holder Number 2": Can't compute!
{{CONTEXT_HOLDER_ID(Test Holder Number 2): CONTEXT_HOLDER_INFO(Will fail)}:[
(index(0), depth(0), ID(<function random_int at 0x000001D1D8A4EE18>), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BUILT-IN EXCEPTION: ChainLinkFailed (Too small!)))),
(index(1), depth(0), ID(<function random_int at 0x000001D1D8A4EE18>), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT))),
(index(2), depth(0), ID(<function int_div at 0x000001D1D8A4EEA0>), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT))),
(index(3), depth(0), ID(Result), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT)))
]}

ERROR IN "Test Holder Number 3": Can't compute!
{{CONTEXT_HOLDER_ID(Test Holder Number 3): CONTEXT_HOLDER_INFO(Every function should return number >= 0. Will fail.)}:[
(index(0), depth(0), ID(<function random_int at 0x000001D1D8A4EE18>), INFO(None), RESULT((True, 100))),
(index(1), depth(0), ID(<function random_int at 0x000001D1D8A4EE18>), INFO(None), RESULT((False, -5))),
(index(2), depth(0), ID(<function int_div at 0x000001D1D8A4EEA0>), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT))),
(index(3), depth(0), ID(Result), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT)))
]}

ERROR IN "Test Holder Number 4": Can't compute!
{{CONTEXT_HOLDER_ID(Test Holder Number 4): CONTEXT_HOLDER_INFO(Every function should return even number. Will fail.)}:[
(index(0), depth(0), ID(<function random_int at 0x000001D1D8A4EE18>), INFO(None), RESULT((True, 100))),
(index(1), depth(0), ID(<function random_int at 0x000001D1D8A4EE18>), INFO(None), RESULT((True, 4))),
(index(2), depth(0), ID(<function int_div at 0x000001D1D8A4EEA0>), INFO(None), RESULT((False, 25.0))),
(index(3), depth(0), ID(Result), INFO(None), RESULT((False, ISOK INTERNAL RESULT. BLOCK DID NOT RETURN RESULT)))
]}


========================

    :return:
    """
    # ============================
    # TEST 1:
    cont = Chain('Test Holder Number 1', 'Will be successful', ResultType(CriteriaType.optional, set()))
    cr = ChainUniCallRunner(cont, True)  # Simple-Mode - for any functions

    # Will be successful.
    # Will return 625 if you'll try to assign final result to the variable
    cr('Result')(
        int_square,
        cr()(
            int_div,
            cr()(random_int, 100, 100),  # 100
            cr()(random_int, 4, 4)  # 4
        )  # 25
    )  # 625

    # Will print "625"
    ChainReaderCallRunner(cont)()(print_result, 'Result')

    # ============================
    print()

    # ============================
    # TEST 2:
    cont = Chain('Test Holder Number 2', 'Will fail', ResultType(CriteriaType.optional, set()))
    cr = ChainUniCallRunner(cont, True)  # Simple-Mode - for any functions

    # Will fail: -40 < -20.
    # Will return None if you'll try to assign final result to the variable
    cr('Result')(
        int_square,  # Will not be computed
        cr()(
            int_div,  # Will not be computed
            cr()(random_int, -40, 0),  # -40 < -20. EXCEPTION ChainLinkFailed ("Too small!")
            cr()(random_int, 3, 4)))  # Will not be computed

    # Will print an error
    ChainReaderCallRunner(cont)()(print_result, 'Result')

    # ============================
    print()

    # ============================
    # TEST 3:
    cont = Chain('Test Holder Number 3', 'Every function should return number >= 0. Will fail.',
                 ResultType(CriteriaType.optional, set()))
    cr = ChainUniCallRunner(cont, True, lambda result: result >= 0)  # Simple-Mode - for any functions

    # Will fail: random_int(-5, -1) < 0.
    # Will return None if you'll try to assign final result to the variable
    cr('Result')(
        int_square,
        cr()(
            int_div,
            cr()(random_int, 100, 100),  # 100
            cr()(random_int, -5, -1)  # result < 0. Result state is FAIL (push_result(False, ...))
        )  # Will not be computed
    )  # Will not be computed

    # Will print an error
    ChainReaderCallRunner(cont)()(print_result, 'Result')

    # ============================
    print()

    # ============================
    # TEST 4:
    cont = Chain('Test Holder Number 4', 'Every function should return even number. Will fail.',
                 ResultType(CriteriaType.optional, set()))
    cr = ChainUniCallRunner(cont, True, lambda result: result % 2 == 0)  # Simple-Mode - for any functions

    # Will fail: 25 is odd.
    # Will return None if you'll try to assign final result to the variable
    int_1 = cr()(random_int, 100, 100)  # 100
    int_2 = cr()(random_int, 4, 4)  # 4
    div_res = cr()(int_div, int_1, int_2)  # 25. Result is odd. Result state is FAIL (push_result(False, 25))
    cr('Result')(int_square, div_res)  # Will not be computed

    # Will print an error
    ChainReaderCallRunner(cont)()(print_result, 'Result')


def main():
    separator = '========================\n'
    title_template = separator + '=== {} ===' + '\n'
    block_separator = ('\n' * 2) + (separator * 3) + ('\n' * 2)

    print(title_template.format('WITH-CHAIN-RAISE-EXCEPTIONS:'))
    with_chain_example(True)

    print(block_separator)
    print(title_template.format('WITH-CHAIN:'))
    with_chain_example(False)

    print(block_separator)
    print(title_template.format('FUNC-DECORATOR-COMPLICATED-CHAIN:'))
    func_decorator_chain_example()

    print(block_separator)
    print(title_template.format('FUNC-RUNNER-COMPLICATED-CHAIN:'))
    func_runner_chain_complicated_example()

    print(block_separator)
    print(title_template.format('FUNC-RUNNER-SIMPLE-CHAIN:'))
    func_runner_chain_simple_example()

    print(block_separator)
    print(title_template.format('FUNC-CALL-RUNNER-COMPLICATED-CHAIN:'))
    func_call_runner_chain_complicated_example()

    print(block_separator)
    print(title_template.format('FUNC-CALL-RUNNER-SIMPLE-CHAIN:'))
    func_call_runner_chain_simple_example()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'OpenTextFile',
    'TextFileInfo',
]


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.text_processing.encoding_detection import detect_and_decode
from cengal.code_flow_control.smart_values import ValueHolder

from pathlib import PurePath
from os import PathLike
from io import IOBase
from typing import Union, Tuple, Optional, NamedTuple


class TextFileInfo(NamedTuple):
    text: ValueHolder[str]
    encoding: str
    bom_bytes: bytes
    bin_data: bytes
    file: IOBase


class OpenTextFile:
    def __init__(self, 
                file: Union[str, bytes, PathLike],
                mode: str = 'r+b',
                buffering: int = -1,
                encoding = None,
                errors = None,
                newline = None,
                closefd = True,
                opener = None,
                detect_as_utf8_when_possible: bool = True,
                check_text_for_utf8_compliance: bool = True,
                ):
        self.file: Union[str, bytes, PathLike] = file
        self.mode: str = mode
        self.buffering = buffering
        self.encoding = encoding
        self.proposed_encoding = encoding
        self.errors = errors
        self.newline = newline
        self.closefd = closefd
        self.opener = opener
        self.detect_as_utf8_when_possible: bool = detect_as_utf8_when_possible
        self.check_text_for_utf8_compliance: bool = check_text_for_utf8_compliance
        self.readable: bool = ('r' in mode) or ('+' in mode)
        self.writtable: bool = ('w' in mode) or ('a' in mode) or ('+' in mode)
        self.binary_mode: bool = 'b' in mode
        if self.binary_mode:
            self.encoding = None
        
        self.text_value_holder: Optional[ValueHolder[str]] = None
        self.detected_encoding: str = None
        self.bom_bytes: bytes = None
    
    def __enter__(self) -> TextFileInfo:
        if (not self.readable) and (not self.proposed_encoding):
            raise ValueError('Encoding must be specified for files without reading mode')

        self.file = open(self.file, self.mode, self.buffering, self.encoding, self.errors, self.newline, self.closefd, self.opener)
        if self.binary_mode:
            if self.readable:
                bin_data: bytes = self.file.read()
                text, encoding, bom_bytes = detect_and_decode(bin_data, self.detect_as_utf8_when_possible, self.check_text_for_utf8_compliance)
            else:
                bin_data = bytes()
                text = str()
                encoding = self.proposed_encoding
                bom_bytes = bytes()
            
            self.detected_encoding = encoding
            self.bom_bytes = bom_bytes
            self.text_value_holder = ValueHolder(self.writtable, text)
            return TextFileInfo(self.text_value_holder, encoding, bom_bytes, bin_data, self.file)
        else:
            encoding = self.proposed_encoding
            if self.readable:
                text: str = self.file.read()
                bin_text: bytes = text.encode(encoding)
            else:
                bin_text = bytes()
                text = str()
            
            self.detected_encoding = encoding
            bom_bytes = bytes()
            self.bom_bytes = bom_bytes
            self.text_value_holder = ValueHolder(self.writtable, text)
            return TextFileInfo(self.text_value_holder, encoding, bom_bytes, bin_text, self.file)
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.writtable and self.text_value_holder:
            self.file.write(self.bom_bytes + self.text_value_holder.value.encode(self.detected_encoding))
        
        self.file.close()
        return False
    
    async def __aenter__(self) -> TextFileInfo:
        # TODO: implement async version using next backends in exact priority sequence: 
        # ['https://github.com/mosquito/aiofile', 'https://github.com/Tinche/aiofiles', 'own asyncio thread based read-write implementation']. 
        # Dependencies must be optional so own implementation is a last viable option.
        return self.__enter__()
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return self.__exit__(exc_type, exc_val, exc_tb)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import colorama
from cengal.user_interface.console.colorama_helpers import colorama_init

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


_LINE_TEMPLATE = '   {}: {}{}{}'


class Chooser:
    def __init__(self, list_of_options: list, variant_type_name: str=None, line_template: str=None):
        self.list_of_options = list_of_options
        self.variant_type_name = variant_type_name or 'variant'
        self.line_template = line_template or _LINE_TEMPLATE

    def choose(self, last_tool_id_str: str=None):
        # last_tool_id_str = last_tool_id_str or ''
        input_raw_last_tool_id = str()
        tool_number = None
        tool_is_chosen = False

        with colorama_init():
            print('{}Chose {} By Entering Number:{}'.format(colorama.Fore.YELLOW, self.variant_type_name.title(),
                                                            colorama.Style.RESET_ALL))
            index = 1
            for tool_name in self.list_of_options:
                print(self.line_template.format(
                    tool_name,
                    colorama.Fore.GREEN,
                    index,
                    colorama.Style.RESET_ALL
                ))
                index += 1

            while not tool_is_chosen:
                try:
                    if last_tool_id_str is None:
                        tool_number = input('Chose {0} ID number: '.format(
                                self.variant_type_name))
                    else:
                        tool_number = input('Chose {0} ID number or leave empty to use last used {0} ID ({1}): '.format(
                            self.variant_type_name, last_tool_id_str))
                    if len(tool_number) == 0:
                        tool_number = last_tool_id_str
                    if (tool_number is None) or len(tool_number) == 0:
                        tool_number = '-1'
                    tool_number = int(tool_number)
                    if (tool_number < 1) or (tool_number > len(self.list_of_options)):
                        raise ValueError
                    input_raw_last_tool_id = str(tool_number)
                    tool_number -= 1
                    tool_is_chosen = True
                except ValueError:
                    pass
            print()

        result = (tool_is_chosen, tool_number, input_raw_last_tool_id)
        return result



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'CengalBuildExtension',
    'CengalSetuptoolsBuildExtension',
    'CengalCythonBuildExtension',
    'CengalGeneratorBuildExtension',
]


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


# from distutils.dist import Distribution
from os import environ

from setuptools._distutils.dist import Distribution

from cengal.file_system.path_manager import path_relative_to_src, RelativePath, get_relative_path_part, sep
from cengal.file_system.directory_manager import current_src_dir, change_current_dir
from cengal.file_system.directory_manager import filtered_file_list, FilteringType, filtered_file_list_traversal, file_list_traversal, FilteringEntity
from cengal.file_system.file_manager import current_src_file_dir, file_exists
from cengal.build_tools.prepare_cflags import prepare_cflags, concat_cflags, prepare_compile_time_env, adjust_definition_names, \
    dict_of_tuples_to_dict, list_to_dict
from cengal.introspection.inspect import get_exception, entity_repr_limited_try_qualname, pifrl, pdi
from cengal.text_processing.text_processing import find_text
from cengal.system import OS_TYPE, TEMPLATE_MODULE_NAME
from shutil import rmtree
from os import remove
from os.path import splitext, normpath, join as path_join, basename, split
from setuptools import Extension as SetuptoolsExtension
from Cython.Distutils import Extension as CythonExtension
from distutils.command.build import build as build_orig
from distutils.command.build_ext import build_ext as build_ext_orig
from setuptools.command.sdist import sdist as sdist_orig
import json
import importlib

from os.path import isdir, exists, isfile, dirname

import setuptools
import platform

from cengal.file_system.path_manager import RelativePath, get_relative_path_part
from cengal.file_system.directory_manager import current_src_dir
from cengal.file_system.directory_manager import file_list_traversal, FilteringEntity
from cengal.build_tools.prepare_cflags import prepare_compile_time_flags, prepare_compile_time_env
from cengal.os.execute import prepare_params, escape_text, escape_param, prepare_command
from setuptools.discovery import find_package_path
import subprocess
from pprint import pprint
from typing import List, Dict, Optional, Iterable, Callable, Sequence, Tuple, Union, Type, Any


class CengalBuildExtension:
    base_class: Optional[Type] = None
    store_as_data: bool = False

    def __init__(self, kwargs) -> None:
        self.kwargs = kwargs
        self._path = None
        self._module_import_str = None
        self._files = list()

    def __call__(self) -> Optional[List[str]]:
        return self.base_class(**self.kwargs) if self.base_class is not None else None
    
    def sdist() -> Optional[List[str]]:
        raise NotImplementedError
    
    @property
    def path(self) -> str:
        return self._path

    @path.setter
    def path(self, value: str):
        self._path = value
    
    @property
    def dir_path(self) -> str:
        return None if self._path is None else dirname(self._path)
    
    @property
    def dir_path_rel(self) -> RelativePath:
        return None if self._path is None else RelativePath(dirname(self._path))

    @property
    def module_import_str(self) -> str:
        return self._module_import_str
    
    @module_import_str.setter
    def module_import_str(self, value: str):
        self._module_import_str = value

    @property
    def name(self) -> str:
        return self._module_import_str

    @property
    def package(self) -> str:
        return '.'.join(self._module_import_str.split('.')[:-1])
    
    @property
    def files(self) -> str:
        return self._files
    
    @files.setter
    def files(self, value: str):
        self._files = value


class CengalSetuptoolsBuildExtension(CengalBuildExtension):
    base_class: Optional[Type] = SetuptoolsExtension

    def __init__(self, **kwargs) -> None:
        super().__init__(kwargs)


class CengalCythonBuildExtension(CengalBuildExtension):
    base_class: Optional[Type] = CythonExtension

    def __init__(self, **kwargs) -> None:
        super().__init__(kwargs)


class CengalGeneratorBuildExtension(CengalBuildExtension):
    base_class: Optional[Type] = None
    store_as_data: bool = True

    def __init__(self, **kwargs) -> None:
        super().__init__(kwargs)
    
    def __call__(self):
        raise NotImplementedError



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['prepare_loop', 'prepare_fast_loop']


from asyncio import AbstractEventLoop
from asyncio import Task as asyncio_Task
from asyncio import get_event_loop, get_running_loop

from cengal.code_flow_control.smart_values import ValueExistence
from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import (
    PutCoro, PutCoroRequest)
from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import \
    RunCoro
from cengal.parallel_execution.coroutines.coro_standard_services.wait_coro import (
    CoroutineNotFoundError, WaitCoro, WaitCoroRequest)
from cengal.parallel_execution.coroutines.coro_standard_services.instance import Instance, InstanceRequest
from cengal.parallel_execution.coroutines.coro_tools.await_coro import (
    RunSchedulerInAsyncioLoop, await_coro_fast)
from cengal.time_management.sleep_tools import get_usable_min_sleep_interval
from typing import Tuple, Any, Union, Optional

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class LoopWasEndedBeforeSetupWasPrepared(Exception):
    pass


def prepare_loop(setup_coro_worker: Optional[AnyWorker], *args, **kwargs) -> Tuple[CoroSchedulerType, ValueExistence[Any]]:
    cs = CoroScheduler()
    set_primary_coro_scheduler(cs)
    cs.turn_on_embedded_mode(True)
    cs.set_coro_time_measurement(True)
    cs.set_coro_history_gathering(True)
    cs.set_loop_iteration_time_measurement(True)
    cs.set_global_on_start_handlers(True)
    cs.suppress_coro_exceptions = True
    cs.suppress_warnings_about_responses_to_not_existant_coroutines = True
    cs.use_internal_sleep = True

    async def initial_setup(i: Interface):
        await i(PutCoro, PutCoroRequest().turn_on_tree_monitoring(True))

    initial_setup_coro: CoroWrapperBase = cs.put_coro(initial_setup)

    async def main(i: Interface, initial_setup_coro: CoroWrapperBase, setup_coro_worker: Optional[AnyWorker], *args, **kwargs):
        try:
            await i(WaitCoro, WaitCoroRequest().single(initial_setup_coro.coro_id))
        except CoroutineNotFoundError:
            pass

        if setup_coro_worker is None:
            return None
        
        return await i(RunCoro, setup_coro_worker, *args, **kwargs)

    coro: CoroWrapperBase = cs.put_coro(main, initial_setup_coro, setup_coro_worker, *args, **kwargs)
    result: ValueExistence = ValueExistence()

    def gather(coro: CoroWrapperBase) -> bool:
        result.value = (coro.last_result, coro.exception)
        return True

    coro.add_on_coro_del_handler(gather)
    in_work: bool = True
    while in_work and (not result):
        in_work = cs.iteration()
    
    if not result:
        raise LoopWasEndedBeforeSetupWasPrepared
    
    result, exception = result.value
    if exception is not None:
        raise exception

    return cs, result


def prepare_fast_loop(setup_coro_worker: Optional[AnyWorker], *args, **kwargs) -> Tuple[CoroSchedulerType, ValueExistence[Any]]:
    """30% faster than prepare_loop()

    Args:
        coro_worker (AnyWorker): _description_

    Raises:
        exception.value: _description_

    Returns:
        _type_: _description_
    """
    cs = CoroScheduler()
    set_primary_coro_scheduler(cs)
    cs.turn_on_embedded_mode(False)
    cs.set_coro_time_measurement(False)
    cs.set_coro_history_gathering(False)
    cs.set_loop_iteration_time_measurement(False)
    cs.set_global_on_start_handlers(False)
    cs.suppress_coro_exceptions = True
    cs.suppress_warnings_about_responses_to_not_existant_coroutines = True
    cs.use_internal_sleep = True

    async def initial_setup(i: Interface):
        await i(PutCoro, PutCoroRequest().turn_on_tree_monitoring(False))

    initial_setup_coro: CoroWrapperBase = cs.put_coro(initial_setup)

    async def main(i: Interface, initial_setup_coro: CoroWrapperBase, setup_coro_worker: Optional[AnyWorker], *args, **kwargs):
        try:
            await i(WaitCoro, WaitCoroRequest().single(initial_setup_coro.coro_id))
        except CoroutineNotFoundError:
            pass

        if setup_coro_worker is None:
            return None

        return await i(RunCoro, setup_coro_worker, *args, **kwargs)

    coro: CoroWrapperBase = cs.put_coro(main, initial_setup_coro, setup_coro_worker, *args, **kwargs)
    result: ValueExistence = ValueExistence()

    def gather(coro: CoroWrapperBase) -> bool:
        result.value = (coro.last_result, coro.exception)
        return True

    coro.add_on_coro_del_handler(gather)
    in_work: bool = True
    while in_work and (not result):
        in_work = cs.iteration()
    
    if not result:
        raise LoopWasEndedBeforeSetupWasPrepared
    
    result, exception = result.value
    if exception is not None:
        raise exception

    return cs, result



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import sys

from typing import Tuple
from square_and_rectangle import *


def check_for_exit(input_str: str) -> None:
    if input_str.lower() in {'q', 'quit', 'exit'}:
        raise SystemExit


def get_new_x_y() -> Tuple[bool, float, float]:
    is_ok = True
    x = 0.0
    y = 0.0
    try:
        input_str = input('Enter a new x y dimension: ')
        check_for_exit(input_str)
        input_str_list = input_str.strip().split()
        x = float(input_str_list[0])
        y = float(input_str_list[1])
    except EssenceModelException:
        is_ok = False
    return is_ok, x, y


def get_stretch() -> Tuple[bool, float]:
    is_ok = True
    multiplier = 0.0
    try:
        input_str = input('SQUARE stretch multiplier: ')
        check_for_exit(input_str)
        multiplier = float(input_str.strip())
    except EssenceModelException:
        is_ok = False
    return is_ok, multiplier


def get_x_or_y_stretch() -> Tuple[bool, bool, float]:
    is_ok = True
    is_x = True
    multiplier = 0.0
    try:
        input_str = input('RECTANGLE stretch in a format "is_x multiplier" ("yes 4"): ')
        check_for_exit(input_str)
        input_str_list = input_str.strip().split()
        if input_str_list[0].lower() in {'y', 'yes'}:
            is_x = True
        else:
            is_x = False
        multiplier = float(input_str_list[1])
    except EssenceModelException:
        is_ok = False
    return is_ok, is_x, multiplier


def request_dimension_callback(rectangle: RectangleInterface):
    print('Current dimension: ', rectangle.dimension().x, rectangle.dimension().y)
    is_ok, new_x, new_y = get_new_x_y()
    if is_ok:
        rectangle.set_dimension(Point(new_x, new_y))


def stretch_square_callback(stretching_square: StretchingSquareInterface):
    print('Current dimension: ', stretching_square.dimension().x, stretching_square.dimension().y)
    is_ok, multiplier = get_stretch()
    if is_ok:
        stretching_square.stretch(multiplier)


def stretch_rectangle_callback(stretching_rectangle: StretchingRectangleInterface):
    print('Current dimension: ', stretching_rectangle.dimension().x, stretching_rectangle.dimension().y)
    is_ok, is_x, multiplier = get_x_or_y_stretch()
    if is_ok:
        if is_x:
            stretching_rectangle.stretch_x(multiplier)
        else:
            stretching_rectangle.stretch_y(multiplier)


def main():
    frame: RectangleModel = rectangle_factory(Point(0, 0))
    while True:
        frame(RectangleInterface, request_dimension_callback)
        frame(StretchingSquareInterface, stretch_square_callback)  # Will be executed only if frame is a square
        frame(StretchingRectangleInterface, stretch_rectangle_callback)
        frame(StretchingSquareInterface, stretch_square_callback)  # Will be executed only if frame is a square


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['get_min_sleep_interval', 'get_usable_min_sleep_interval', 'get_countable_delta_time',
           'ensure_waitable_time_or_smaller', 'ensure_waitable_time_or_bigger', 'try_sleep', 'sleep', 'atry_sleep']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"

import os
from typing import Optional, Union

from cengal.math.numbers import RationalNumber
from cengal.code_flow_control.args_manager import EntityArgsHolder
from cengal.time_management.high_precision_sync_sleep import hps_sleep
from cengal.system import IS_INSIDE_OR_FOR_WEB_BROWSER

if IS_INSIDE_OR_FOR_WEB_BROWSER:
    default_sleep = hps_sleep
else:
    from time import sleep as default_sleep


if IS_INSIDE_OR_FOR_WEB_BROWSER:
    _default_min_sleep_interval = 0.001
elif 'nt' == os.name:
    _default_min_sleep_interval = 0.0156
else:
    _default_min_sleep_interval = 0.0001


_countable_delta_time = 0.0000001


def get_min_sleep_interval():
    return _default_min_sleep_interval


def get_usable_min_sleep_interval():
    return _default_min_sleep_interval + _countable_delta_time


def get_countable_delta_time():
    return _countable_delta_time


def ensure_waitable_time_or_smaller(secs: RationalNumber) -> RationalNumber:
    secs = abs(secs)
    intervals: int = secs // _default_min_sleep_interval
    if not intervals:
        return 0

    return _default_min_sleep_interval * intervals + _countable_delta_time


def ensure_waitable_time_or_bigger(secs: RationalNumber) -> RationalNumber:
    if secs < get_usable_min_sleep_interval():
        return 0

    intervals: int = secs // _default_min_sleep_interval
    if secs % _default_min_sleep_interval:
        intervals += 1

    return _default_min_sleep_interval * intervals + _countable_delta_time


def try_sleep(secs: RationalNumber, max_secs: Optional[RationalNumber], sleep_func, *args, **kwargs):
    result = None
    usable_min_sleep_interval = get_usable_min_sleep_interval()

    if max_secs is not None:
        if usable_min_sleep_interval > max_secs:
            max_secs = usable_min_sleep_interval

        if secs > max_secs:
            secs = max_secs

    secs = _default_min_sleep_interval * (secs // _default_min_sleep_interval) + _countable_delta_time
    if secs >= _default_min_sleep_interval:
        still_trying = True
        while still_trying:
            secs = _default_min_sleep_interval * (secs // _default_min_sleep_interval) + _countable_delta_time

            if secs < usable_min_sleep_interval:
                secs = usable_min_sleep_interval

            try:
                # print(f'Sleepting: {secs}...')
                result = sleep_func(secs, *args, **kwargs)
                # print(f'Awaken!')
            except OverflowError:
                still_trying = True
                secs /= 2
            else:
                still_trying = False

    return result, secs


def sleep(secs: Optional[RationalNumber] = None, high_cpu_utilisation_mode: bool = False, max_secs: Optional[RationalNumber] = None, sleep_func: Optional[EntityArgsHolder] = None):
    result = None
    usable_min_sleep_interval = get_usable_min_sleep_interval()
    if secs is None:
        secs = usable_min_sleep_interval

    if max_secs is not None:
        if usable_min_sleep_interval > max_secs:
            max_secs = usable_min_sleep_interval

        if secs > max_secs:
            secs = max_secs

    secs = _default_min_sleep_interval * (secs // _default_min_sleep_interval) + _countable_delta_time
    if secs >= _default_min_sleep_interval:
        still_trying = True
        while still_trying:
            secs = _default_min_sleep_interval * (secs // _default_min_sleep_interval) + _countable_delta_time

            if secs < usable_min_sleep_interval:
                secs = usable_min_sleep_interval

            try:
                # print(f'Sleepting: {secs}...')
                if sleep_func is None:
                    result = default_sleep(secs)
                else:
                    sleep_func, args, kwargs = sleep_func.entity_args_kwargs()
                    result = sleep_func(secs, *args, **kwargs)
                # print(f'Awaken!')
            except OverflowError:
                still_trying = True
                secs /= 2
            else:
                still_trying = False
    else:
        if high_cpu_utilisation_mode and (not IS_INSIDE_OR_FOR_WEB_BROWSER):
            result = hps_sleep(secs)

    return result, secs


async def atry_sleep(secs: RationalNumber, max_secs: Optional[RationalNumber], awaitable_sleep_func, *args, **kwargs):
    result = None
    usable_min_sleep_interval = get_usable_min_sleep_interval()

    if max_secs is not None:
        if usable_min_sleep_interval > max_secs:
            max_secs = usable_min_sleep_interval

        if secs > max_secs:
            secs = max_secs

    secs = _default_min_sleep_interval * (secs // _default_min_sleep_interval) + _countable_delta_time
    if secs >= _default_min_sleep_interval:
        still_trying = True
        while still_trying:
            secs = _default_min_sleep_interval * (secs // _default_min_sleep_interval) + _countable_delta_time

            if secs < usable_min_sleep_interval:
                secs = usable_min_sleep_interval

            try:
                result = await awaitable_sleep_func(secs, *args, **kwargs)
            except OverflowError:
                still_trying = True
                secs /= 2
            else:
                still_trying = False

    return result, secs



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['run_coroutine_in_new_thread', 'ProcessPoolRuntimeError', 'ExecutorSetupBase', 'ExecutorTypeSetup', 'ExecutorInstanceSetup', 'InitializerSetup', 'ProcessPoolSetup', 'ProcessPool']

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.introspection.inspect import get_exception, is_async
from cengal.code_flow_control.smart_values import ResultHolder

import sys
import asyncio
from concurrent.futures import Executor
from threading import Thread
from functools import partial
from typing import Callable, Optional, Type, Any, Tuple


class ProcessPoolRuntimeError(RuntimeError):
    pass


class ExecutorSetupBase:
    pass


class ExecutorTypeSetup(ExecutorSetupBase):
    def __init__(self, executor_type: Type[Executor], *args, **kwargs) -> None:
        self.executor_type: Type[Executor] = executor_type
        self.args = args
        self.kwargs = kwargs
        super().__init__()


class ExecutorInstanceSetup(ExecutorSetupBase):
    def __init__(self, executor: Executor) -> None:
        self.executor: Executor = executor
        super().__init__()


class InitializerSetup:
    def __init__(self, initializer: Callable, *args, **kwargs) -> None:
        self.initializer: Callable = initializer
        self.args = args
        self.kwargs = kwargs


class ProcessPoolSetup:
    def __init__(self, is_multiprocessing: bool = True, loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        self.loop: Optional[asyncio.AbstractEventLoop] = loop or asyncio.get_event_loop()
        self.is_multiprocessing: bool = is_multiprocessing


def run_coroutine_in_new_thread(coro: Callable, *args, **kwargs) -> Any:
    def thread_worker(result_holder: ResultHolder, coro: Callable, *args, **kwargs):
        result = None
        exception = None
        try:
            import asyncio
            result = asyncio.run(coro(*args, **kwargs))
        except:
            exception = get_exception()
        
        result_holder.value = (result, exception)
    
    result_holder: ResultHolder = ResultHolder()
    thread_args = [result_holder, coro,] + list(args)
    thread: Thread = Thread(target=thread_worker, args=thread_args, kwargs=kwargs)
    thread.start()
    thread.join()
    if result_holder:
        result, exception = result_holder.value
    else:
        result = None
        exception = RuntimeError('ProcessPool internal error')
    
    return result, exception


class ProcessPool:
    def __init__(self, executor_setup: ExecutorSetupBase, initializer_setup: Optional[InitializerSetup] = None, process_pool_setup: Optional[ProcessPoolSetup] = None) -> None:
        self.executor_setup: ExecutorSetupBase = executor_setup
        self.process_pool_setup: Optional[ProcessPoolSetup] = process_pool_setup or ProcessPoolSetup()
        self.initializer_setup: Optional[InitializerSetup] = initializer_setup
        self.executor: Executor = None
        self.partial_pool_initializer: Callable = None
        
        self._create_pool()
    
    def set_is_multiprocessing(self, is_multiprocessing: bool):
        self.process_pool_setup.is_multiprocessing = is_multiprocessing
    
    def _create_pool(self):
        if self.initializer_setup is not None:
            self.partial_pool_initializer = partial(ProcessPool._initializer, self.initializer_setup.initializer, *self.initializer_setup.args, **self.initializer_setup.kwargs)
        
        if (3, 7) <= sys.version_info:
            if isinstance(self.executor_setup, ExecutorInstanceSetup):
                self.executor = self.executor_setup.executor
            elif isinstance(self.executor_setup, ExecutorTypeSetup):
                if self.partial_pool_initializer is not None:
                    self.executor_setup.kwargs['initializer'] = self.partial_pool_initializer
                
                self.executor = self.executor_setup.executor_type(*self.executor_setup.args, **self.executor_setup.kwargs)
            else:
                raise ProcessPoolRuntimeError('Unknown "executor_setup" parameter type')
        else:
            self.executor = self.executor_setup.executor_type(*self.executor_setup.args, **self.executor_setup.kwargs)
    
    def shutdown(self, wait=True, cancel_futures=False):
        self.executor.shutdown(wait, cancel_futures=cancel_futures)
    
    @staticmethod
    def _initializer(initializer, *args, **kwargs):
        import multiprocessing
        
        initializer(multiprocessing.current_process()._identity, *args, **kwargs)

    @staticmethod
    def _pool_worker(worker: Callable, *args, **kwargs) -> Tuple[Any, Optional[BaseException]]:
        result = None
        exception = None
        try:
            result = worker(*args, **kwargs)
        except:
            exception = get_exception()
        
        return result, exception

    @staticmethod
    def _pool_worker_wrapper(partial_pool_initializer: Callable, worker: Callable, *args, **kwargs) -> Tuple[Any, Optional[BaseException]]:
        global process_initialized
        if 'process_initialized' not in globals():
            process_initialized = False
        
        if not process_initialized:
            if partial_pool_initializer is not None:
                partial_pool_initializer()
            
            process_initialized = True
        
        return ProcessPool._pool_worker(worker, *args, **kwargs)

    @staticmethod
    def _apool_worker(result_holder: ResultHolder, worker: Callable, *args, **kwargs) -> Tuple[Any, Optional[BaseException]]:
        result = None
        exception = None
        try:
            import asyncio
            result = asyncio.run(worker(*args, **kwargs))
        except:
            exception = get_exception()
        
        result_holder.value = (result, exception)
    
    @staticmethod
    def _apool_worker(worker: Callable, *args, **kwargs) -> Tuple[Any, Optional[BaseException]]:
        return run_coroutine_in_new_thread(worker, *args, **kwargs)

    @staticmethod
    def _apool_worker_wrapper(partial_pool_initializer: Callable, worker: Callable, *args, **kwargs) -> Tuple[Any, Optional[BaseException]]:
        global process_initialized
        if 'process_initialized' not in globals():
            process_initialized = False
        
        if not process_initialized:
            if partial_pool_initializer is not None:
                partial_pool_initializer()

            process_initialized = True
        
        return ProcessPool._apool_worker(worker, *args, **kwargs)

    @staticmethod
    async def _a_single_process_pool_worker(worker: Callable, *args, **kwargs) -> Tuple[Any, Optional[BaseException]]:
        result = None
        exception = None
        try:
            result = await worker(*args, **kwargs)
        except:
            exception = get_exception()
        
        return result, exception

    @staticmethod
    async def _a_single_process_pool_worker_wrapper(partial_pool_initializer: Callable, worker: Callable, *args, **kwargs) -> Tuple[Any, Optional[BaseException]]:
        global process_initialized
        if 'process_initialized' not in globals():
            process_initialized = False
        
        if not process_initialized:
            if partial_pool_initializer is not None:
                partial_pool_initializer()

            process_initialized = True
        
        return await ProcessPool._a_single_process_pool_worker(worker, *args, **kwargs)

    async def pool_execute(self, worker: Callable, *args, **kwargs) -> Any:
        if self.process_pool_setup.is_multiprocessing:
            if ((3, 7) <= sys.version_info) or (self.partial_pool_initializer is None):
                if is_async(worker):
                    partial_pool_worker = partial(ProcessPool._apool_worker, worker, *args, **kwargs)
                else:
                    partial_pool_worker = partial(ProcessPool._pool_worker, worker, *args, **kwargs)
            else:
                if is_async(worker):
                    partial_pool_worker = partial(ProcessPool._apool_worker_wrapper, self.partial_pool_initializer, worker, *args, **kwargs)
                else:
                    partial_pool_worker = partial(ProcessPool._pool_worker_wrapper, self.partial_pool_initializer, worker, *args, **kwargs)
            
            result: Tuple[Any, Optional[BaseException]] = await self.process_pool_setup.loop.run_in_executor(self.executor, partial_pool_worker)
        else:
            if is_async(worker):
                a_single_process_pool_worker = partial(ProcessPool._a_single_process_pool_worker_wrapper, self.partial_pool_initializer, worker, *args, **kwargs)
                result = await a_single_process_pool_worker()
            else:
                partial_pool_worker = partial(ProcessPool._pool_worker_wrapper, self.partial_pool_initializer, worker, *args, **kwargs)
                result = partial_pool_worker()
        
        result, exception = result
        
        if exception is not None:
            raise exception
        
        return result
    
    async def __call__(self, worker: Callable, *args, **kwargs) -> Any:
        return await self.pool_execute(worker, *args, **kwargs)


# def pool_initializer(text):
#     import multiprocessing
    
#     print(text, multiprocessing.current_process()._identity[0])


# def create_pool():
#     global executor_init_params
#     executor_init_params = (('hello pool',), dict())
#     global executor
#     if (3, 7) <= sys.version_info:
#         pool_args, pool_kwargs = executor_init_params
#         partial_pool_initializer = partial(pool_initializer, *pool_args, **pool_kwargs)
#         executor = ProcessPoolExecutor(max_workers=2, initializer=partial_pool_initializer)
#     else:
#         executor = ProcessPoolExecutor(max_workers=2)


# def pool_worker_impl(item: int):
#     return 1000 / item


# def pool_worker(*args, **kwargs):
#     result = None
#     exception = None
#     try:
#         result = pool_worker_impl(*args, **kwargs)
#     except:
#         exception = get_exception()
    
#     return result, exception


# def pool_worker_wrapper(pool_init_params, *args, **kwargs):
#     global process_initialized
#     if 'process_initialized' not in globals():
#         process_initialized = False
    
#     if not process_initialized:
#         pool_args, pool_kwargs = pool_init_params
#         pool_initializer(*pool_args, **pool_kwargs)
#         process_initialized = True
    
#     return pool_worker(*args, **kwargs)
        

# async def pool_execute(*args, **kwargs):
#     if (3, 7) <= sys.version_info:
#         partial_pool_worker = partial(pool_worker, *args, **kwargs)
#     else:
#         partial_pool_worker = partial(pool_worker_wrapper, executor_init_params, *args, **kwargs)
        
#     if is_multiprocessing:
#         result = await loop.run_in_executor(executor, partial_pool_worker)
#     else:
#         result = pool_worker(*args, **kwargs)
    
#     result, exception = result
    
#     if exception is not None:
#         raise exception
    
#     return result


# async def pool_single_processing_example(item: int = 2):
#     return await pool_execute(item)


# async def pool_gather_example(num: int = 3):
#     return await asyncio.gather(*[pool_execute(i) for i in range(num)])



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.unittest.behavior_stabilizer import UnittestTestCaseWithState, UnittestTestCaseState
from cengal.performance_test_lib import PrecisePerformanceTestTracer
from cengal.introspection.inspect import func_name
from threading import Lock, RLock
from pprint import pprint
from typing import Callable


def func_perf_test(func: Callable, *args, **kwargs):
    tr = PrecisePerformanceTestTracer(1.0)
    while tr.iter():
        func(*args, **kwargs)

    with tr as fast_iter:
        for i in fast_iter:
            func(*args, **kwargs)

    print('{}(): {} iter/s; {} seconds; {} iterations'.format(func_name(func), tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))


from cengal.hardware.memory.shared_memory import *
from cengal.time_management.cpu_clock import cpu_clock
from cengal.time_management.high_precision_sync_sleep import hps_sleep
from multiprocessing import Process
from multiprocessing.resource_tracker import unregister
import _posixshmem

try:
    shm_name = '/test_shmem'
    _posixshmem.shm_unlink(shm_name)
    unregister(shm_name, "shared_memory")
except FileNotFoundError:
    pass

import unittest


# class TestCaseForRunInLoop(UnittestTestCaseWithState):
class TestCaseForRunInLoop(unittest.TestCase):
    # @classmethod
    # def setUpClass(cls):
    #     cls.state = UnittestTestCaseState()

    # @classmethod
    # def tearDownClass(cls):
    #     cls.state.unregister()
    
    def test_shared_memory(self):
        results: List[Any] = list()
        creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

        try:
            results.append(creator.read_mem(creator.get_data_start_offset(), 50))
            offset, size = creator.malloc(ObjectType.tbool, 1)
            results.append(size)
            results.append(creator.read_mem(creator.get_data_start_offset(), 50))
            offset, size = creator.realloc(offset, 11)
            results.append(size)
            results.append(creator.read_mem(creator.get_data_start_offset(), 50))
            results.append(creator.free(offset))
            results.append(creator.read_mem(creator.get_data_start_offset(), 50))
            creator.set_free_memory_search_start(0)
            creator.commit_free_memory_search_start()
            results.append(creator.read_mem(creator.get_data_start_offset(), 50))
            offset, size = creator.malloc(ObjectType.tbool, 1)
            results.append(size)
            results.append(creator.read_mem(creator.get_data_start_offset(), 50))
            offset, size = creator.realloc(offset, 11)
            results.append(size)
            results.append(creator.read_mem(creator.get_data_start_offset(), 50))
            results.append(creator.free(offset))
            results.append(creator.read_mem(creator.get_data_start_offset(), 50))
            creator.set_free_memory_search_start(0)
            creator.commit_free_memory_search_start()

            values = [
                0,
                1,
                999999999999,
                -1,
                -999999999999,
                0.0,
                1.0,
                999999999999.0,
                -1.0,
                -999999999999.0,
                (1,),
                [1],
                (1, 2),
                [1, 2],
                (1, [2, 3]),
                {1: 1, 2: 2},
                {1: 1, 2: [2, (3,)]},
                'test',
                'test'.encode('utf-8'),
                bytearray('test'.encode('utf-8')),
                True,
                False,
                {
                    'test': 1,
                    'test2': 2.0,
                    'test3': 'test',
                    'test4': [
                        1, 
                        2.0, 
                        True, 
                        False, 
                        'test', 
                        'test'.encode('utf-8'), 
                        bytearray('test'.encode('utf-8')), 
                        ('hello', 1),
                        ['world', 2.0],
                        {
                            'hello': [1, (2, 3)],
                            'world': 'world',
                        }
                    ],
                }
            ]
            
            results.append(creator.read_mem(creator.get_free_memory_search_start(), 100))
            results.append(creator.read_mem(creator.get_data_start_offset(), 500))
            
            for val in values:
                creator.put_message(val)
                results.append(creator.read_mem(creator.get_free_memory_search_start(), 100))
                results.append(creator.read_mem(creator.get_data_start_offset(), 500))
                
                message = creator.take_message()
                self.assertEqual(val, message)
                results.append(message)
                results.append(creator.read_mem(creator.get_free_memory_search_start(), 100))
                results.append(creator.read_mem(creator.get_data_start_offset(), 500))
        finally:
            creator._shared_memory.close()
            creator._shared_memory.unlink()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.code_flow_control.smart_values.versions.v_2 import *
from unittest import TestCase, main


class TestSmartValues(TestCase):
    def test__value_existence(self):
        ve_classess = (
            ValueExistence,
            ValueHolder,
            ResultExistence,
            ErrorExistence,
        )
        for ve_class in ve_classess:
            value_existence = ve_class()
            self.assertEqual(value_existence.existence, False)
            self.assertEqual(value_existence[0], False)
            self.assertEqual(value_existence.value, None)
            self.assertEqual(value_existence[1], None)
            self.assertEqual(tuple(value_existence), (False, None))
            self.assertEqual(tuple(value_existence), value_existence.to_tuple())
            self.assertEqual(tuple(ve_class.from_tuple((True, 30))), (True, 30))

        for ve_class in ve_classess:
            value_existence = ve_class(True, 10)
            self.assertEqual(value_existence.existence, True)
            self.assertEqual(value_existence[0], True)
            self.assertEqual(value_existence.value, 10)
            self.assertEqual(value_existence[1], 10)
            self.assertEqual(tuple(value_existence), (True, 10))
            self.assertEqual(tuple(value_existence), value_existence.to_tuple())
            self.assertEqual(tuple(ve_class.from_tuple((True, 30))), (True, 30))

        for ve_class in ve_classess:
            value_existence = ve_class[int]()
            self.assertEqual(value_existence.existence, False)
            self.assertEqual(value_existence[0], False)
            self.assertEqual(value_existence.value, None)
            self.assertEqual(value_existence[1], None)
            self.assertEqual(tuple(value_existence), (False, None))
            self.assertEqual(tuple(value_existence), value_existence.to_tuple())
            self.assertEqual(tuple(ve_class.from_tuple((True, 30))), (True, 30))

        for ve_class in ve_classess:
            value_existence = ve_class[int](True, 10)
            self.assertEqual(value_existence.existence, True)
            self.assertEqual(value_existence[0], True)
            self.assertEqual(value_existence.value, 10)
            self.assertEqual(value_existence[1], 10)
            self.assertEqual(tuple(value_existence), (True, 10))
            self.assertEqual(tuple(value_existence), value_existence.to_tuple())
            self.assertEqual(tuple(ve_class.from_tuple((True, 30))), (True, 30))
            
        value_cache = ValueCache()
        self.assertFalse(value_cache)
        self.assertEqual(value_cache.existence, False)
        self.assertEqual(value_cache[0], False)
        self.assertEqual(value_cache.get(), None)
        self.assertEqual(value_cache.value, None)
        self.assertEqual(value_cache[1], None)
        self.assertEqual(tuple(value_cache), (False, None))
        self.assertEqual(tuple(value_cache), value_cache.to_tuple())
        value_cache.set(10)
        self.assertTrue(value_cache)
        self.assertEqual(value_cache.existence, True)
        self.assertEqual(value_cache[0], True)
        self.assertEqual(value_cache.get(), 10)
        self.assertEqual(value_cache.value, 10)
        self.assertEqual(value_cache[1], 10)
        self.assertEqual(tuple(value_cache), (True, 10))
        self.assertEqual(tuple(value_cache), value_cache.to_tuple())
        self.assertEqual(tuple(ValueCache.from_tuple((True, 30))), (True, 30))
            
        value_cache = ValueCache[int]()
        self.assertFalse(value_cache)
        self.assertEqual(value_cache.existence, False)
        self.assertEqual(value_cache[0], False)
        self.assertEqual(value_cache.get(), None)
        self.assertEqual(value_cache.value, None)
        self.assertEqual(value_cache[1], None)
        self.assertEqual(tuple(value_cache), (False, None))
        self.assertEqual(tuple(value_cache), value_cache.to_tuple())
        value_cache.set(10)
        self.assertTrue(value_cache)
        self.assertEqual(value_cache.existence, True)
        self.assertEqual(value_cache[0], True)
        self.assertEqual(value_cache.get(), 10)
        self.assertEqual(value_cache.value, 10)
        self.assertEqual(value_cache[1], 10)
        self.assertEqual(tuple(value_cache), (True, 10))
        self.assertEqual(tuple(value_cache), value_cache.to_tuple())
        self.assertEqual(tuple(ValueCache[int].from_tuple((True, 30))), (True, 30))

        value_type = ValueType(4, 'hello')
        self.assertEqual(value_type.type_id, 4)
        self.assertEqual(value_type[0], 4)
        self.assertEqual(value_type.value, 'hello')
        self.assertEqual(value_type[1], 'hello')
        self.assertEqual(tuple(value_type), (4, 'hello'))
        self.assertEqual(tuple(value_type), value_type.to_tuple())
        self.assertEqual(tuple(ValueType.from_tuple(('nullable', 30))), ('nullable', 30))

        value_type = ValueType[int, str](4, 'hello')
        self.assertEqual(value_type.type_id, 4)
        self.assertEqual(value_type[0], 4)
        self.assertEqual(value_type.value, 'hello')
        self.assertEqual(value_type[1], 'hello')
        self.assertEqual(tuple(value_type), (4, 'hello'))
        self.assertEqual(tuple(value_type), value_type.to_tuple())
        self.assertEqual(tuple(ValueType[str, int].from_tuple(('nullable', 30))), ('nullable', 30))

        value_with_type = ValueWithType(4, 'hello')
        self.assertEqual(value_with_type.type_id, 4)
        self.assertEqual(value_with_type[0], 4)
        self.assertEqual(value_with_type.value, 'hello')
        self.assertEqual(value_with_type[1], 'hello')
        self.assertEqual(tuple(value_with_type), (4, 'hello'))
        self.assertEqual(tuple(value_with_type), value_with_type.to_tuple())
        self.assertEqual(tuple(ValueWithType.from_tuple(('nullable', 30))), ('nullable', 30))

        value_with_type = ValueWithType[int, str](4, 'hello')
        self.assertEqual(value_with_type.type_id, 4)
        self.assertEqual(value_with_type[0], 4)
        self.assertEqual(value_with_type.value, 'hello')
        self.assertEqual(value_with_type[1], 'hello')
        self.assertEqual(tuple(value_with_type), (4, 'hello'))
        self.assertEqual(tuple(value_with_type), value_with_type.to_tuple())
        self.assertEqual(tuple(ValueWithType[str, int].from_tuple(('nullable', 30))), ('nullable', 30))



if '__main__' == __name__:
    main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from cengal.code_flow_control.args_manager import ArgsKwargs, AK
from collections import namedtuple
from typing import Hashable, Dict, Tuple, List, Any, AnyStr, Union, Sequence

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


ItemTemplate = namedtuple("ItemTemplate", "l_delimiter, prefix, template, postfix, r_delimiter", defaults=('', '', '{}', '', ''))
IT = ItemTemplate


class OptionalFormatter:
    def __init__(self, item_positions: Tuple[Hashable, ...],
                 template_per_item: Dict[Hashable, Union[IT, Tuple[str, str, str, str, str]]]):
        self._item_positions = item_positions
        self._template_per_item = template_per_item

    def __call__(self, arguments_per_item: Dict[Hashable, Union[AK, Any]]) -> str:
        result = str()
        is_first = True
        last_postfix = str()
        last_r_delimiter = str()
        for item in self._item_positions:
            if item in arguments_per_item:
                item_l_delimiter, item_prefix, item_template, item_postfix, item_r_delimiter = \
                    self._template_per_item[item]
                item_args_kwargs = arguments_per_item[item]
                if isinstance(item_args_kwargs, AK):
                    item_args, item_kwargs = arguments_per_item[item]()
                else:
                    item_args = (item_args_kwargs,)
                    item_kwargs = dict()
                
                rendered_item = str()
                if is_first:
                    rendered_item += item_prefix
                else:
                    rendered_item += last_r_delimiter
                    rendered_item += item_l_delimiter
                
                last_r_delimiter = item_r_delimiter
                last_postfix = item_postfix
                rendered_item += item_template.format(*item_args, **item_kwargs)
                result += rendered_item
                if is_first:
                    is_first = False
                
        result += last_postfix
        return result


class OptionalFormatterHandy(OptionalFormatter):
    """
    template_per_item second tuple format:
    [left_delimiter(shown if not first), prefix(shown if first), template, postfix(shown if last), right_delimiter(shown if not last)]


    Example 1:
        f = OptionalFormatterHandy('hour', 'part_of_day', 'minute', 'second', 'millisecond',
                                   hour=IT('', '|(hours)', '{}', '|'),
                                   part_of_day=IT('-', '|(part_of_day)', '<{}.{second}>', '|'),
                                   minutes=IT(':', '|(minutes)', '{}', '|'),
                                   seconds=IT(':', '|(seconds)', '{}', '|'),
                                   millisecond=IT('.', '|(millisecond)', '{}', '|'))

        f(hour=4, minutes=15, seconds=54, millisecond=341)
        >> |(hours)4:15:54.341|

        f(seconds=54, minutes=15, hour=0, millisecond=0)
        >> |(hours)0:15:54.0|

        f(minutes=15, seconds=54)
        >> |(minutes)15:54|

        f(hour=4, part_of_day=AK('a', second='m'), minutes=15, seconds=54, millisecond=341)
        >> |(hours)4-<a.m>:15:54.341|

        f(part_of_day=AK('a', second='m'), minutes=15, seconds=54, millisecond=341)
        >> |(part_of_day)<a.m>:15:54.341|

        
    Example 2:
        f = OptionalFormatterHandy('hour', 'word', 'minute', 'second', 'millisecond',
                                   hour=('', '|(hours)', '{}', '|', ''),
                                   word=('', '|(word)', '-<"{}"-"{second}">-', '|', ''),
                                   minutes=(':', '|(minutes)', '{}', '|', ''),
                                   seconds=(':', '|(seconds)', '{}', '|', ''),
                                   millisecond=('.', '|(millisecond)', '{}', '|', ''))

        f(hour=4, minutes=15, seconds=54, millisecond=341)
        >> |(hours)4:15:54.341|

        f(seconds=54, minutes=15, hour=0, millisecond=0)
        >> |(hours)0:15:54.0|

        f(minutes=15, seconds=54)
        >> |(minutes)15:54|

        f(hour=4, word=AK('HELLO!', second='WORLD!'), minutes=15, seconds=54, millisecond=341)
        >> |(hours)4-<"HELLO!"-"WORLD!">-:15:54.341|

        f(word=AK('HELLO!', second='WORLD!'), minutes=15, seconds=54, millisecond=341)
        >> |(word)-<"HELLO!"-"WORLD!">-:15:54.341|
    """
    def __init__(self, *item_positions: [str, ...], **template_per_item: [str, Union[IT, Tuple[str, str, str, str, str]]]):
        super().__init__(item_positions, template_per_item)

    def __call__(self, **arguments_per_item: [str, Union[AK, Any]]) -> str:
        return super(OptionalFormatterHandy, self).__call__(arguments_per_item)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['sync_coro', 'sync_coro_param']


from cengal.parallel_execution.coroutines.coro_scheduler import *
# from cengal.parallel_execution.coroutines.coro_standard_services import *
from typing import Union, Optional, Any, List, Tuple, Dict
from cengal.parallel_execution.coroutines.coro_standard_services_internal_lib.service_with_a_direct_request import *
from cengal.code_flow_control.smart_values import ValueExistence
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import put_coro_to
from cengal.parallel_execution.coroutines.coro_standard_services.wait_coro import WaitCoro, WaitCoroRequest
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro_list import PSCP
from cengal.parallel_execution.coroutines.coro_tools.prepare_loop import prepare_loop, prepare_fast_loop


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class LoopWasEndedBeforeResultWasComputed(Exception):
    pass


def sync_coro_param(result_required: bool = True, timeout: Optional[float] = None, kill_on_timeout: bool = True, tree: bool = True, cs: Optional[CoroSchedulerType] = None, prepare_own_loop_if_not_found: bool = False, own_loop_shold_be_fast_loop: bool = True, own_loop_setup_coro_worker: Optional[AnyWorker] = None, own_loop_setup_coro_worker_args_kwargs: Optional[Tuple[Tuple, Dict]] = None):
    """Decorator. With an arguments. Gives ability to execute any decorated Cengal coroutine from plain sync/async 
    code as a sync function. Can postpone execution to the actual loop when possible if None as an immediate result 
    (no result) is acceptible. Can start own loop if needed.

    Args:
        result_required (bool, optional): _description_. Defaults to True.
        timeout (Optional[float], optional): _description_. Defaults to None.
        kill_on_timeout (bool, optional): _description_. Defaults to True.
        tree (bool, optional): _description_. Defaults to True.
        cs (Optional[CoroSchedulerType], optional): _description_. Defaults to None.
        prepare_own_loop_if_not_found (bool, optional): _description_. Defaults to False.
        own_loop_shold_be_fast_loop (bool, optional): _description_. Defaults to True.
        own_loop_setup_coro_worker (Optional[AnyWorker], optional): _description_. Defaults to None.
        own_loop_setup_coro_worker_args_kwargs (Optional[Tuple[Tuple, Dict]], optional): _description_. Defaults to None.

    Raises:
        exception: _description_
        exception: _description_
        LoopWasEndedBeforeResultWasComputed: _description_
        TypeError: _description_
        TypeError: _description_

    Returns:
        _type_: _description_
    """    
    cs0 = cs
    def sync_coro_decorator(coro_worker: Worker):
        cs1 = cs0
        def wrapper(*args, **kwargs):
            cs = cs1
            done: ValueExistence = ValueExistence()
            async def coro(i: Interface, done: ValueExistence, coro_worker: Worker, *args, **kwargs):
                results: List[Tuple[CoroID, Any, Union[None, Exception]]] = await i(WaitCoro, WaitCoroRequest(timeout, kill_on_timeout, tree).put_list([PSCP(coro_worker, *args, **kwargs)]))
                done.value = results[0]
            
            direct_execution_required: bool = False
            if cs is None:
                cs = get_available_coro_scheduler()
            
            if (cs is None) and prepare_own_loop_if_not_found:
                if own_loop_setup_coro_worker_args_kwargs is None:
                    args: Tuple = tuple()
                    kwargs: Dict = dict()
                else:
                    args, kwargs = own_loop_setup_coro_worker_args_kwargs

                if own_loop_shold_be_fast_loop:
                    cs, _ = prepare_fast_loop(own_loop_setup_coro_worker, *args, **kwargs)
                else:
                    cs, _ = prepare_loop(own_loop_setup_coro_worker, *args, **kwargs)

            if cs is None:
                direct_execution_required = True
            else:
                i: Optional[Interface] = None
                inside_loop: bool = False
                inside_coro: bool = False
                try:
                    i = current_interface()
                    inside_loop = True
                    if i is not None:
                        inside_coro = True
                except OutsideCoroSchedulerContext:
                    pass

                if inside_loop:
                    if inside_coro:
                        inside_greenlet_coro: bool = False
                        if isinstance(i, InterfaceGreenlet):
                            inside_greenlet_coro = True
                        
                        if inside_greenlet_coro:
                            results: List[Tuple[CoroID, Any, Union[None, Exception]]] = i(WaitCoro, WaitCoroRequest(timeout, kill_on_timeout, tree).put_list([PSCP(coro_worker, *args, **kwargs)]))
                            _, result, exception = results[0]
                            if exception is not None:
                                raise exception
                            
                            return result
                        else:
                            pass
                    else:
                        pass
                else:
                    put_coro_to(get_interface_and_loop_with_explicit_loop(cs), coro, done, coro_worker, *args, **kwargs)
                    in_work = True
                    while in_work and (not done):
                        in_work = cs.iteration()
                    
                    if done:
                        _, result, exception = done.value
                        if exception is not None:
                            raise exception
                    else:
                        raise LoopWasEndedBeforeResultWasComputed

            if result_required:
                direct_execution_required = True
            else:
                put_coro_to(interface_and_loop_with_explicit_loop(cs), coro_worker, *args, **kwargs)

            if direct_execution_required:
                coro_worker_type: CoroType = find_coro_type(coro_worker)
                if CoroType.greenlet == coro_worker_type:
                    return coro_worker(InterfaceFake(None, None), *args, **kwargs)
                elif CoroType.awaitable == coro_worker_type:
                    raise TypeError(f'Can not directly execute awaitable {coro_worker} from non async environment')
                else:
                    raise TypeError(f'{coro_worker} is neither an awaitable nor a greenlet')
        
        return wrapper
    return sync_coro_decorator


def sync_coro(coro_worker: Worker):
    return sync_coro_param()(coro_worker)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['FrontTriggerableVariableType', 'FrontTriggerableVariable']


from enum import Enum
from typing import Union

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class FrontTriggerableVariableType(Enum):
    equal = 0
    lesser = 1
    lesser_or_equal = 2
    bigger = 3
    bigger_or_equal = 4
    not_equal = 5


class FrontTriggerableVariable:
    def __init__(self, triggerable_variable_type, value_limit):
        self.triggerable_variable_type = triggerable_variable_type
        self.value_limit = value_limit
        self._last_result = None
        self.test_worker = None
        self.set_triggerable_variable_type(triggerable_variable_type)
    
    def set_triggerable_variable_type(self, triggerable_variable_type):
        self._last_result = None
        self.triggerable_variable_type = triggerable_variable_type
        if FrontTriggerableVariableType.equal == triggerable_variable_type:
            self.test_worker = FrontTriggerableVariable._equal
        elif FrontTriggerableVariableType.lesser == triggerable_variable_type:
            self.test_worker = FrontTriggerableVariable._lesser
        elif FrontTriggerableVariableType.lesser_or_equal == triggerable_variable_type:
            self.test_worker = FrontTriggerableVariable._lesser_or_equal
        elif FrontTriggerableVariableType.bigger == triggerable_variable_type:
            self.test_worker = FrontTriggerableVariable._bigger
        elif FrontTriggerableVariableType.bigger_or_equal == triggerable_variable_type:
            self.test_worker = FrontTriggerableVariable._bigger_or_equal
        elif FrontTriggerableVariableType.not_equal == triggerable_variable_type:
            self.test_worker = FrontTriggerableVariable._not_equal

    def test_trigger(self, value) -> Union[None, bool]:
        result = None
        new_test_result = self.test_worker(value, self.value_limit)
        if new_test_result != self._last_result:
            self._last_result = new_test_result
            result = new_test_result
        
        return result

    def __call__(self, value) -> Union[None, bool]:
        return self.test_trigger(value)

    @staticmethod
    def _equal(value, value_limit) -> bool:
        return value == value_limit

    @staticmethod
    def _lesser(value, value_limit) -> bool:
        return value < value_limit

    @staticmethod
    def _lesser_or_equal(value, value_limit) -> bool:
        return value <= value_limit

    @staticmethod
    def _bigger(value, value_limit) -> bool:
        return value > value_limit

    @staticmethod
    def _bigger_or_equal(value, value_limit) -> bool:
        return value >= value_limit

    @staticmethod
    def _not_equal(value, value_limit) -> bool:
        return value != value_limit



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cengal.code_flow_control.smart_values.versions.v_0 import ResultExistence
import string
from cengal.data_manipulation.conversion.binary import ubyte_to_bytes
from cengal.data_manipulation.conversion.sequence import get_slice_from_array
from typing import Tuple, Union

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class AbstractSetOfSymbols:
    def __init__(self):
        self.ascii_word_delimiters = None
        self.ascii_word_delimiters__bytes = None
        self.ascii_word_delimiters__set = None
        self.ascii_word_delimiters__set_bytes = None

        # without "_" symbol
        self.ascii_modern_word_delimiters = None
        self.ascii_modern_word_delimiters__bytes = None
        self.ascii_modern_word_delimiters__set = None
        self.ascii_modern_word_delimiters__set_bytes = None


class SetOfSymbols(AbstractSetOfSymbols):
    def __init__(self):
        super(SetOfSymbols, self).__init__()
        self.ascii_word_delimiters = string.punctuation + string.whitespace
        self.ascii_word_delimiters__bytes = self.ascii_word_delimiters.encode()
        self.ascii_word_delimiters__set = set(self.ascii_word_delimiters)
        self.ascii_word_delimiters__set_bytes = set()
        for delim_char in self.ascii_word_delimiters__set:
            self.ascii_word_delimiters__set_bytes.add(delim_char.encode())

        self.ascii_modern_word_delimiters__set = set(self.ascii_word_delimiters__set)
        self.ascii_modern_word_delimiters__set.remove('_')
        self.ascii_modern_word_delimiters = ''.join(self.ascii_modern_word_delimiters__set)
        self.ascii_modern_word_delimiters__bytes = self.ascii_modern_word_delimiters.encode()
        self.ascii_modern_word_delimiters__set_bytes = set()
        for delim_char in self.ascii_modern_word_delimiters__set:
            self.ascii_modern_word_delimiters__set_bytes.add(delim_char.encode())


SET_OF_SYMBOLS = SetOfSymbols()


def get_text_in_brackets(data, left_b, right_b):
    # TODO: если в строке не найдена закрывающая скобка - последний символ строки будет удален. Проверить, не ломает ли
    # такое поведение алгоритмы в частности в upk-утилитах и в UCB-компиляторе
    left_offset = data.find(left_b)
    data = data[left_offset + len(left_b):]
    right_offset = data.find(right_b)
    data = data[:right_offset]
    return data


def get_text_in_brackets_offset(data, left_b, right_b, offset=0):
    # TODO: если в строке не найдена закрывающая скобка - последний символ строки будет удален. Проверить, не ломает ли
    # такое поведение алгоритмы в частности в upk-утилитах и в UCB-компиляторе
    result = None
    result_data = None
    result_offset = None
    if offset > 0:
        data = data[offset:]
    left_b_len = len(left_b)
    right_b_len = len(right_b)
    left_offset = data.find(left_b)
    data = data[left_offset + left_b_len:]
    right_offset = data.find(right_b)
    result_data = data[:right_offset]
    result_offset = offset + left_offset + left_b_len + right_offset + right_b_len
    result = (result_data, result_offset)
    return result


def detach_slice_from_string(string, substring, offset=0):
    slice_index = string.index(substring, offset)
    slice_size = len(substring)
    end_of_slice = slice_index + slice_size
    string_before = string[:slice_index]
    string_slice = string[slice_index:end_of_slice]
    string_after = string[end_of_slice:]
    result = (string_before, string_slice, string_after, end_of_slice)
    return result


def detach_slice_from_string__case_insensitive(string, substring, offset=0):
    lowercase_string = string.lower()
    lowercase_substring = substring.lower()
    slice_index = lowercase_string.index(lowercase_substring, offset)
    slice_size = len(substring)
    end_of_slice = slice_index + slice_size
    string_before = string[:slice_index]
    string_slice = string[slice_index:end_of_slice]
    string_after = string[end_of_slice:]
    result = (string_before, string_slice, string_after, end_of_slice)
    return result


def find_substring(full_string: bytes, substring: bytes, offset: int=0)\
        ->Tuple[Union[None, int], Union[None, int]]:
    start_index = full_string.find(substring, offset)
    word_start = True
    if 0 > start_index:
        word_start = False

    end_index = start_index + len(substring)

    if not word_start:
        start_index = None
        end_index = None

    return start_index, end_index


def find_substring_full_word(full_string: bytes, substring: bytes, offset: int=0, smart_word_bounds: bool=False)\
        ->Tuple[ResultExistence, ResultExistence]:
    start_index = ResultExistence(False, 0)
    end_index = ResultExistence(False, None)

    while start_index.result is not None:
        start_index, end_index = find_substring_full_word__one_shot(full_string, substring, offset, smart_word_bounds)
        if start_index.result is not None:
            # substring was found
            if start_index and end_index:
                # full word was found
                break
            else:
                # need to continue search from the new offset
                offset = end_index.result
        else:
            # substring wasn't found
            break

    if start_index:
        start_index = start_index.result
    else:
        start_index = None

    if end_index:
        end_index = end_index.result
    else:
        end_index = None

    return start_index, end_index


class FindSubstringErrorFullStringCanNotBeEmpty(Exception):
    pass


class FindSubstringErrorSubstringCanNotBeEmpty(Exception):
    pass


def find_substring_full_word__one_shot(full_string: bytes, substring: bytes, offset: int=0,
                                       smart_word_bounds: bool=False)\
        ->Tuple[ResultExistence, ResultExistence]:
    delimiters = SET_OF_SYMBOLS.ascii_modern_word_delimiters__set_bytes

    word_start = ResultExistence(False, None)
    word_end = ResultExistence(False, None)

    if not full_string:
        return word_start, word_end
    if not substring:
        return word_start, word_end

    is_word_start_is_delimiter = False
    is_word_end_is_delimiter = False
    if smart_word_bounds:
        if ubyte_to_bytes(substring[0]) in delimiters:
            is_word_start_is_delimiter = True
        if ubyte_to_bytes(substring[-1]) in delimiters:
            is_word_end_is_delimiter = True

    word_start = ResultExistence(False, full_string.find(substring, offset))
    word_end = ResultExistence(False, None)

    if 0 > word_start.result:
        word_start.existence = False
        word_start.result = None
    else:
        if is_word_start_is_delimiter:
            word_start.existence = True
        elif 0 == word_start.result:
            word_start.existence = True
        elif ubyte_to_bytes(full_string[word_start.result - 1]) in delimiters:
            word_start.existence = True

    if word_start.result is not None:
        word_end.result = word_start.result + len(substring)
        full_string_len = len(full_string)
        if is_word_end_is_delimiter:
            word_start.existence = True
        elif word_end.result > full_string_len:
            word_end.existence = False
        elif word_end.result == full_string_len:
            word_end.existence = True
        elif ubyte_to_bytes(full_string[word_end.result]) in delimiters:
            word_end.existence = True

    return word_start, word_end


def check_is_slice_is_in_string(substring: bytes, full_string: bytes, check_whole_word=False,
                                smart_word_bounds: bool=False):
    if check_whole_word:
        word_start, word_end = find_substring_full_word(full_string, substring, smart_word_bounds=smart_word_bounds)
        if (word_start is not None) and (word_end is not None):
            return True
        else:
            return False
    else:
        return substring in full_string


def check_is_slice_is_in_string__case_insensitive(substring: bytes, full_string: bytes, check_whole_word=False,
                                                  smart_word_bounds: bool=False):
    return check_is_slice_is_in_string(substring.lower(), full_string.lower(), check_whole_word, smart_word_bounds)


def detach_all_slices_from_string(string, substring, function__detach=None, function__check_is_in=None):
    """
    :param string: input string
    :param substring: desired substring
    :param function__detach: detach_slice_from_string (when None) or detach_slice_from_string__case_insensitive
    :param function__check_is_in: check_is_slice_is_in_string (when None) or
        check_is_slice_is_in_string__case_insensitive
    :return: ([(original_string_part_0, string_slice), (original_string_part_1, string_slice), ...,
        (original_string_part_N, string_slice)], string_after)
    """
    function__detach = function__detach or detach_slice_from_string
    function__check_is_in = function__check_is_in or check_is_slice_is_in_string

    result_list = list()
    result = function__detach(string, substring)
    last_string_after = result[2]
    new_result = (result[0], result[1])
    result_list.append(new_result)
    while function__check_is_in(substring, result[2]):
        result = function__detach(result[2], substring)
        last_string_after = result[2]
        new_result = (result[0], result[1])
        result_list.append(new_result)
    result = (result_list, last_string_after)
    return result


def detach_all_slices_from_string__case_insensitive(string, substring):
    return detach_all_slices_from_string(
        string, substring, detach_slice_from_string__case_insensitive, check_is_slice_is_in_string__case_insensitive)


def is_printable(s, codec='utf8'):
    try:
        s.decode(codec)
    except UnicodeDecodeError:
        return False
    else: 
        return True


def bytes_to_printable(bytes_data):
    result = str(bytes_data)[2:-1]
    return result


def levenshtein_distance(a, b):
    "Calculates the Levenshtein distance between a and b."
    n, m = len(a), len(b)
    if n > m:
        # Make sure n <= m, to use O(min(n,m)) space
        a, b = b, a
        n, m = m, n

    current_row = range(n+1) # Keep current and previous row, not entire matrix
    for i in range(1, m+1):
        previous_row, current_row = current_row, [i]+[0]*n
        for j in range(1,n+1):
            add, delete, change = previous_row[j]+1, current_row[j-1]+1, previous_row[j-1]
            if a[j-1] != b[i-1]:
                change += 1
            current_row[j] = min(add, delete, change)

    return current_row[n]


def un_escape_str(text: str) -> str:
    return bytes(text, 'utf-8').decode('unicode_escape')
